@@ -1,8 +1,26 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
+#
 # Copyright (C) 2004 Robert Kaye
-# Copyright (C) 2006-2007 Lukáš Lalinský
+# Copyright (C) 2006-2009, 2011-2012, 2014 Lukáš Lalinský
+# Copyright (C) 2008 Gary van der Merwe
+# Copyright (C) 2008 Hendrik van Antwerpen
+# Copyright (C) 2008 ojnkpjg
+# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009 Nikolai Prokoschenko
+# Copyright (C) 2011-2012 Chad Wilson
+# Copyright (C) 2011-2013, 2019 Michael Wiencek
+# Copyright (C) 2012-2013, 2016-2017 Wieland Hoffmann
+# Copyright (C) 2013, 2018 Calvin Walton
+# Copyright (C) 2013-2015, 2017 Sophist-UK
+# Copyright (C) 2013-2015, 2017-2019 Laurent Monin
+# Copyright (C) 2016 Suhas
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2017 Antonio Larrosa
+# Copyright (C) 2018 Vishal Choudhary
+# Copyright (C) 2019 Joel Lintunen
+# Copyright (C) 2020 Gabriel Ferreira
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -18,7 +36,12 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-from operator import itemgetter
+
+from collections import (
+    OrderedDict,
+    defaultdict,
+    namedtuple,
+)
 import traceback
 
 from PyQt5 import (
@@ -33,7 +56,6 @@ from picard import (
 from picard.cluster import Cluster
 from picard.collection import add_release_to_user_collections
 from picard.const import VARIOUS_ARTISTS_ID
-from picard.coverart import coverart
 from picard.dataobj import DataObject
 from picard.file import File
 from picard.mbjson import (
@@ -44,10 +66,13 @@ from picard.mbjson import (
 )
 from picard.metadata import (
     Metadata,
-    register_album_metadata_processor,
     run_album_metadata_processors,
     run_track_metadata_processors,
 )
+from picard.plugin import (
+    PluginFunctions,
+    PluginPriority,
+)
 from picard.script import (
     ScriptError,
     ScriptParser,
@@ -55,6 +80,7 @@ from picard.script import (
 )
 from picard.track import Track
 from picard.util import (
+    find_best_match,
     format_time,
     mbid_validate,
 )
@@ -67,7 +93,16 @@ from picard.util.textencoding import asciipunct
 
 from picard.ui.item import Item
 
-register_album_metadata_processor(coverart)
+
+def _create_artist_node_dict(source_node):
+    return {x['artist']['id']: x['artist'] for x in source_node['artist-credit']}
+
+
+def _copy_artist_nodes(source, target_node):
+    for credit in target_node['artist-credit']:
+        artist_node = source.get(credit['artist']['id'])
+        if artist_node:
+            credit['artist'] = artist_node
 
 
 class AlbumArtist(DataObject):
@@ -90,7 +125,9 @@ class Album(DataObject, Item):
         self._files = 0
         self._requests = 0
         self._tracks_loaded = False
-        self._discid = discid
+        self._discids = set()
+        if discid:
+            self._discids.add(discid)
         self._after_load_callbacks = []
         self.unmatched_files = Cluster(_("Unmatched Files"), special=True, related_album=self, hide_if_empty=True)
         self.errors = []
@@ -119,6 +156,20 @@ class Album(DataObject, Item):
         self._album_artists.append(album_artist)
         return album_artist
 
+    def add_discid(self, discid):
+        if not discid:
+            return
+        self._discids.add(discid)
+        for track in self.tracks:
+            medium_discids = track.metadata.getall('~musicbrainz_discids')
+            track_discids = list(self._discids.intersection(medium_discids))
+            if track_discids:
+                track.metadata['musicbrainz_discid'] = track_discids
+                track.update()
+                for file in track.linked_files:
+                    file.metadata['musicbrainz_discid'] = track_discids
+                    file.update()
+
     def get_album_artists(self):
         """Returns the list of album artists (as AlbumArtist objects)"""
         return self._album_artists
@@ -141,6 +192,14 @@ class Album(DataObject, Item):
                 self.tagger.albums[release_id] = self
                 self.id = release_id
 
+        # Make the release artist nodes available, since they may
+        # contain supplementary data (aliases, tags, genres, ratings)
+        # which aren't present in the release group, track, or
+        # recording artist nodes. We can copy them into those places
+        # wherever the IDs match, so that the data is shared and
+        # available for use in mbjson.py and external plugins.
+        self._release_artist_nodes = _create_artist_node_dict(release_node)
+
         # Get release metadata
         m = self._new_metadata
         m.length = 0
@@ -150,13 +209,11 @@ class Album(DataObject, Item):
         rg.loaded_albums.add(self.id)
         rg.refcount += 1
 
+        _copy_artist_nodes(self._release_artist_nodes, rg_node)
         release_group_to_metadata(rg_node, rg.metadata, rg)
         m.copy(rg.metadata)
         release_to_metadata(release_node, m, album=self)
 
-        if self._discid:
-            m['musicbrainz_discid'] = self._discid
-
         # Custom VA name
         if m['musicbrainz_albumartistid'] == VARIOUS_ARTISTS_ID:
             m['albumartistsort'] = m['albumartist'] = config.setting['va_name']
@@ -229,6 +286,11 @@ class Album(DataObject, Item):
             del self._new_metadata
             del self._new_tracks
             self.update()
+            if not self._requests:
+                self.loaded = True
+                for func, always in self._after_load_callbacks:
+                    if always:
+                        func()
             return
 
         if self._requests > 0:
@@ -236,8 +298,9 @@ class Album(DataObject, Item):
 
         if not self._tracks_loaded:
             artists = set()
-            totalalbumtracks = 0
+            all_media = []
             absolutetracknumber = 0
+
             va = self._new_metadata['musicbrainz_albumartistid'] == VARIOUS_ARTISTS_ID
 
             djmix_ars = {}
@@ -249,10 +312,18 @@ class Album(DataObject, Item):
                 mm.copy(self._new_metadata)
                 medium_to_metadata(medium_node, mm)
                 discpregap = False
+                format = medium_node.get('format')
+                if format:
+                    all_media.append(format)
 
                 for dj in djmix_ars.get(mm["discnumber"], []):
                     mm.add("djmixer", dj)
 
+                if 'discs' in medium_node:
+                    discids = [disc.get('id') for disc in medium_node['discs']]
+                    mm['~musicbrainz_discids'] = discids
+                    mm['musicbrainz_discid'] = list(self._discids.intersection(discids))
+
                 if "pregap" in medium_node:
                     discpregap = True
                     absolutetracknumber += 1
@@ -272,13 +343,17 @@ class Album(DataObject, Item):
                         track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
                         track.metadata['~datatrack'] = "1"
 
-            totalalbumtracks = str(absolutetracknumber)
+            totalalbumtracks = absolutetracknumber
+            self._new_metadata['~totalalbumtracks'] = totalalbumtracks
+            # Generate a list of unique media, but keep order of first appearance
+            self._new_metadata['media'] = " / ".join(list(OrderedDict.fromkeys(all_media)))
 
             for track in self._new_tracks:
                 track.metadata["~totalalbumtracks"] = totalalbumtracks
                 if len(artists) > 1:
                     track.metadata["~multiartist"] = "1"
             del self._release_node
+            del self._release_artist_nodes
             self._tracks_loaded = True
 
         if not self._requests:
@@ -305,6 +380,7 @@ class Album(DataObject, Item):
                 for file in list(track.linked_files):
                     file.move(self.unmatched_files)
             self.metadata = self._new_metadata
+            self.orig_metadata.copy(self.metadata)
             self.tracks = self._new_tracks
             del self._new_metadata
             del self._new_tracks
@@ -322,11 +398,22 @@ class Album(DataObject, Item):
                 },
                 timeout=3000
             )
-            for func in self._after_load_callbacks:
+            for func, always in self._after_load_callbacks:
                 func()
             self._after_load_callbacks = []
+            if self.item.isSelected():
+                self.tagger.window.refresh_metadatabox()
 
     def _finalize_loading_track(self, track_node, metadata, artists, va, absolutetracknumber, discpregap):
+        # As noted in `_parse_release` above, the release artist nodes
+        # may contain supplementary data that isn't present in track
+        # artist nodes. Similarly, the track artists may contain
+        # information which the recording artists don't. Copy this
+        # information across to wherever the artist IDs match.
+        _copy_artist_nodes(self._release_artist_nodes, track_node)
+        _copy_artist_nodes(self._release_artist_nodes, track_node['recording'])
+        _copy_artist_nodes(_create_artist_node_dict(track_node), track_node['recording'])
+
         track = Track(track_node['recording']['id'], self)
         self._new_tracks.append(track)
 
@@ -334,7 +421,8 @@ class Album(DataObject, Item):
         tm = track.metadata
         tm.copy(metadata)
         track_to_metadata(track_node, track)
-        track.metadata["~absolutetracknumber"] = absolutetracknumber
+        tm["~absolutetracknumber"] = absolutetracknumber
+        track.orig_metadata.copy(tm)
         track._customize_metadata()
 
         self._new_metadata.length += tm.length
@@ -342,13 +430,13 @@ class Album(DataObject, Item):
         if va:
             tm["compilation"] = "1"
         else:
-            tm.delete("compilation")
+            del tm["compilation"]
         if discpregap:
             tm["~discpregap"] = "1"
 
         # Run track metadata plugins
         try:
-            run_track_metadata_processors(self, tm, self._release_node, track_node)
+            run_track_metadata_processors(self, tm, track_node, self._release_node)
         except BaseException:
             self.error_append(traceback.format_exc())
 
@@ -375,7 +463,7 @@ class Album(DataObject, Item):
         self._requests = 1
         self.errors = []
         require_authentication = False
-        inc = ['release-groups', 'media', 'recordings', 'artist-credits',
+        inc = ['release-groups', 'media', 'discids', 'recordings', 'artist-credits',
                'artists', 'aliases', 'labels', 'isrcs', 'collections']
         if self.tagger.webservice.oauth_manager.is_authorized():
             require_authentication = True
@@ -392,11 +480,11 @@ class Album(DataObject, Item):
             self.id, self._release_request_finished, inc=inc,
             mblogin=require_authentication, priority=priority, refresh=refresh)
 
-    def run_when_loaded(self, func):
+    def run_when_loaded(self, func, always=False):
         if self.loaded:
             func()
         else:
-            self._after_load_callbacks.append(func)
+            self._after_load_callbacks.append((func, always))
 
     def stop_loading(self):
         if self.load_task:
@@ -419,54 +507,64 @@ class Album(DataObject, Item):
         file.metadata_images_changed.disconnect(self.update_metadata_images)
         remove_metadata_images(self, [file])
 
-    def match_files(self, files, use_recordingid=True):
+    def _match_files(self, files, recordingid=None, threshold=0):
         """Match files to tracks on this album, based on metadata similarity or recordingid."""
+        tracks_cache = defaultdict(lambda: None)
+
+        def build_tracks_cache():
+            for track in self.tracks:
+                tm_recordingid = track.orig_metadata['musicbrainz_recordingid']
+                tm_tracknumber = track.orig_metadata['tracknumber']
+                tm_discnumber = track.orig_metadata['discnumber']
+                for tup in (
+                    (tm_recordingid, tm_tracknumber, tm_discnumber),
+                    (tm_recordingid, tm_tracknumber),
+                    (tm_recordingid, )):
+                    tracks_cache[tup] = track
+
+        SimMatchAlbum = namedtuple('SimMatchAlbum', 'similarity track')
+
         for file in list(files):
             if file.state == File.REMOVED:
                 continue
-            matches = []
-            recordingid = file.metadata['musicbrainz_recordingid']
-            if use_recordingid and mbid_validate(recordingid):
-                matches = self._get_recordingid_matches(file, recordingid)
-            if not matches:
+            # if we have a recordingid to match against, use that in priority
+            recid = recordingid or file.metadata['musicbrainz_recordingid']
+            if recid and mbid_validate(recid):
+                if not tracks_cache:
+                    build_tracks_cache()
+                tracknumber = file.metadata['tracknumber']
+                discnumber = file.metadata['discnumber']
+                track = (tracks_cache[(recid, tracknumber, discnumber)]
+                         or tracks_cache[(recid, tracknumber)]
+                         or tracks_cache[(recid, )])
+                if track:
+                    yield (file, track)
+                    continue
+
+            # try to match by similarity
+            def candidates():
                 for track in self.tracks:
-                    sim = track.metadata.compare(file.orig_metadata)
-                    if sim >= config.setting['track_matching_threshold']:
-                        matches.append((sim, track))
-            if matches:
-                matches.sort(key=itemgetter(0), reverse=True)
-                file.move(matches[0][1])
+                    yield SimMatchAlbum(
+                        similarity=track.metadata.compare(file.orig_metadata),
+                        track=track
+                    )
+                    QtCore.QCoreApplication.processEvents()
+
+            no_match = SimMatchAlbum(similarity=-1, track=self.unmatched_files)
+            best_match = find_best_match(candidates, no_match)
+
+            if best_match.similarity < threshold:
+                yield (file, no_match.track)
             else:
-                file.move(self.unmatched_files)
+                yield (file, best_match.result.track)
 
-    def match_file(self, file, recordingid=None):
-        """Match the file on a track on this album, based on recordingid or metadata similarity."""
-        if file.state == File.REMOVED:
-            return
-        if recordingid is not None:
-            matches = self._get_recordingid_matches(file, recordingid)
-            if matches:
-                matches.sort(key=itemgetter(0), reverse=True)
-                file.move(matches[0][1])
-                return
-        self.match_files([file], use_recordingid=False)
-
-    def _get_recordingid_matches(self, file, recordingid):
-        matches = []
-        tracknumber = file.metadata['tracknumber']
-        discnumber = file.metadata['discnumber']
-        for track in self.tracks:
-            tm = track.orig_metadata
-            if recordingid == tm['musicbrainz_recordingid']:
-                if tracknumber == tm['tracknumber']:
-                    if discnumber == tm['discnumber']:
-                        matches.append((4.0, track))
-                        break
-                    else:
-                        matches.append((3.0, track))
-                else:
-                    matches.append((2.0, track))
-        return matches
+    def match_files(self, files, recordingid=None):
+        """Match and move files to tracks on this album, based on metadata similarity or recordingid."""
+        moves = self._match_files(files,
+                                  recordingid=recordingid,
+                                  threshold=config.setting['track_matching_threshold'])
+        for file, target in moves:
+            file.move(target)
 
     def can_save(self):
         return self._files > 0
@@ -577,8 +675,12 @@ class Album(DataObject, Item):
                 return ''
         elif column == 'artist':
             return self.metadata['albumartist']
+        elif column == 'tracknumber':
+            return self.metadata['~totalalbumtracks']
+        elif column == 'discnumber':
+            return self.metadata['totaldiscs']
         else:
-            return ''
+            return self.metadata[column]
 
     def switch_release_version(self, mbid):
         if mbid == self.id:
@@ -642,3 +744,21 @@ class NatAlbum(Album):
 
     def can_browser_lookup(self):
         return False
+
+
+_album_post_removal_processors = PluginFunctions(label='album_post_removal_processors')
+
+
+def register_album_post_removal_processor(function, priority=PluginPriority.NORMAL):
+    """Registers an album-removed processor.
+    Args:
+        function: function to call after album removal, it will be passed the album object
+        priority: optional, PluginPriority.NORMAL by default
+    Returns:
+        None
+    """
+    _album_post_removal_processors.register(function.__module__, function, priority)
+
+
+def run_album_post_removal_processors(album_object):
+    _album_post_removal_processors.run(album_object)
