@@ -1,7 +1,18 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
-# Copyright (C) 2007 Lukáš Lalinský
+#
+# Copyright (C) 2007, 2014, 2016 Lukáš Lalinský
+# Copyright (C) 2014, 2019-2020 Philipp Wolfer
+# Copyright (C) 2014-2016, 2018-2020 Laurent Monin
+# Copyright (C) 2015 Ohm Patel
+# Copyright (C) 2016 Rahul Raturi
+# Copyright (C) 2016 Wieland Hoffmann
+# Copyright (C) 2016-2017 Frederik “Freso” S. Olesen
+# Copyright (C) 2017 Antonio Larrosa
+# Copyright (C) 2017 Sambhav Kothari
+# Copyright (C) 2018 Bob Swift
+# Copyright (C) 2018 Vishal Choudhary
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,17 +28,21 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
+import builtins
 from collections import OrderedDict
 import os
 
 from PyQt5.QtCore import QStandardPaths
+
 from picard import (
     PICARD_APP_NAME,
     PICARD_ORG_NAME,
+    PICARD_VERSION,
 )
+from picard.const.attributes import MB_ATTRIBUTES
 
 # Install gettext "noop" function in case const.py gets imported directly.
-import builtins
 builtins.__dict__['N_'] = lambda a: a
 
 
@@ -54,20 +69,28 @@ MUSICBRAINZ_OAUTH_CLIENT_SECRET = 'xIsvXbIuntaLuRRhzuazOA'
 CAA_HOST = "coverartarchive.org"
 CAA_PORT = 443
 
+# Prepare documentation URLs
+if PICARD_VERSION.identifier == 'final':
+    DOCS_VERSION = "v{}/".format(PICARD_VERSION.to_string(short=True))
+else:
+    DOCS_VERSION = ""  # points to latest version
+DOCS_LANGUAGE = 'en'
+DOCS_BASE_URL = "https://picard-docs.musicbrainz.org/" + DOCS_VERSION + DOCS_LANGUAGE
+
 # URLs
 PICARD_URLS = {
-    'documentation':    "https://picard.musicbrainz.org/docs/",
-    'troubleshooting':  "https://picard.musicbrainz.org/docs/troubleshooting/",
-    'home':             "https://picard.musicbrainz.org/",
-    'doc_options':      "https://picard.musicbrainz.org/docs/options/",
-    'doc_scripting':    "https://picard.musicbrainz.org/docs/scripting",
-    'plugins':          "https://picard.musicbrainz.org/plugins/",
-    'forum':            "https://community.metabrainz.org/c/picard",
-    'donate':           "https://metabrainz.org/donate",
-    'chromaprint':      "https://acoustid.org/chromaprint#download",
-    'acoustid_apikey':  "https://acoustid.org/api-key",
+    'home':                "https://picard.musicbrainz.org/",
+    'documentation':       DOCS_BASE_URL + '/',
+    'troubleshooting':     DOCS_BASE_URL + '/troubleshooting/troubleshooting.html',
+    'doc_options':         DOCS_BASE_URL + '/config/configuration.html',
+    'doc_scripting':       DOCS_BASE_URL + '/scripting.html',
     'doc_cover_art_types': "https://musicbrainz.org/doc/Cover_Art/Types",
-    'acoustid_track': "https://acoustid.org/track/",
+    'plugins':             "https://picard.musicbrainz.org/plugins/",
+    'forum':               "https://community.metabrainz.org/c/picard",
+    'donate':              "https://metabrainz.org/donate",
+    'chromaprint':         "https://acoustid.org/chromaprint#download",
+    'acoustid_apikey':     "https://acoustid.org/api-key",
+    'acoustid_track':      "https://acoustid.org/track/",
 }
 
 # Various Artists MBID
@@ -78,7 +101,6 @@ SILENCE_TRACK_TITLE = '[silence]'
 DATA_TRACK_TITLE = '[data track]'
 
 # Release formats
-from picard.const.attributes import MB_ATTRIBUTES
 RELEASE_FORMATS = {}
 RELEASE_PRIMARY_GROUPS = {}
 RELEASE_SECONDARY_GROUPS = {}
@@ -94,13 +116,13 @@ for k, v in MB_ATTRIBUTES.items():
         RELEASE_STATUS[v] = v
 
 # Release countries
-from picard.const.countries import RELEASE_COUNTRIES  # noqa: F401 # pylint: disable=unused-import
+from picard.const.countries import RELEASE_COUNTRIES  # noqa: F401,E402 # pylint: disable=unused-import
 
 # List of available user interface languages
-from picard.const.languages import UI_LANGUAGES  # noqa: F401 # pylint: disable=unused-import
+from picard.const.languages import UI_LANGUAGES  # noqa: F401,E402 # pylint: disable=unused-import
 
 # List of alias locales
-from picard.const.locales import ALIAS_LOCALES  # noqa: F401 # pylint: disable=unused-import
+from picard.const.locales import ALIAS_LOCALES  # noqa: F401,E402 # pylint: disable=unused-import
 
 # List of official musicbrainz servers - must support SSL for mblogin requests (such as collections).
 MUSICBRAINZ_SERVERS = [
