@@ -21,6 +21,7 @@
 # Copyright (C) 2016 Suhas
 # Copyright (C) 2016-2017 Sambhav Kothari
 # Copyright (C) 2018 Vishal Choudhary
+# Copyright (C) 2020 Gabriel Ferreira
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -86,6 +87,9 @@ from picard.ui.widgets.tristatesortheaderview import TristateSortHeaderView
 
 
 COLUMN_ICON_SIZE = 16
+COLUMN_ICON_BORDER = 2
+ICON_SIZE = QtCore.QSize(COLUMN_ICON_SIZE+COLUMN_ICON_BORDER,
+                         COLUMN_ICON_SIZE+COLUMN_ICON_BORDER)
 
 
 class BaseAction(QtWidgets.QAction):
@@ -163,19 +167,40 @@ class MainPanel(QtWidgets.QSplitter):
         (N_('Fingerprint status'), '~fingerprint'),
     ]
 
-    TITLE_COLUMN = 0
-    FINGERPRINT_COLUMN = 13
+    _column_indexes = {column[1]: i for i, column in enumerate(columns)}
+
+    TITLE_COLUMN = _column_indexes['title']
+    TRACKNUMBER_COLUMN = _column_indexes['tracknumber']
+    DISCNUMBER_COLUMN = _column_indexes['discnumber']
+    LENGTH_COLUMN = _column_indexes['~length']
+    FINGERPRINT_COLUMN = _column_indexes['~fingerprint']
+
+    NAT_SORT_COLUMNS = [
+        _column_indexes['title'],
+        _column_indexes['album'],
+        _column_indexes['discsubtitle'],
+        _column_indexes['tracknumber'],
+        _column_indexes['discnumber'],
+        _column_indexes['catalognumber'],
+    ]
 
     def __init__(self, window, parent=None):
         super().__init__(parent)
         self.window = window
         self.create_icons()
-        self.views = [FileTreeView(window, self), AlbumTreeView(window, self)]
-        self.views[0].itemSelectionChanged.connect(self.update_selection_0)
-        self.views[1].itemSelectionChanged.connect(self.update_selection_1)
-        self._selected_view = 0
+        self._views = [FileTreeView(window, self), AlbumTreeView(window, self)]
+        self._selected_view = self._views[0]
         self._ignore_selection_changes = False
 
+        def _view_update_selection(view):
+            if not self._ignore_selection_changes:
+                self._ignore_selection_changes = True
+                self._update_selection(view)
+                self._ignore_selection_changes = False
+
+        for view in self._views:
+            view.itemSelectionChanged.connect(partial(_view_update_selection, view))
+
         TreeItem.window = window
         TreeItem.base_color = self.palette().base().color()
         TreeItem.text_color = self.palette().text().color()
@@ -194,9 +219,19 @@ class MainPanel(QtWidgets.QSplitter):
             File.ERROR: interface_colors.get_qcolor('entity_error'),
         })
 
+    def set_processing(self, processing=True):
+        self._ignore_selection_changes = processing
+
+    def tab_order(self, tab_order, before, after):
+        prev = before
+        for view in self._views:
+            tab_order(prev, view)
+            prev = view
+        tab_order(prev, after)
+
     def save_state(self):
         config.persist["splitter_state"] = self.saveState()
-        for view in self.views:
+        for view in self._views:
             view.save_state()
 
     @restore_method
@@ -250,33 +285,23 @@ class MainPanel(QtWidgets.QSplitter):
         ]
         self.icon_plugins = icontheme.lookup('applications-system', icontheme.ICON_SIZE_MENU)
 
-    def update_selection(self, i, j):
-        self._selected_view = i
-        self.views[j].clearSelection()
-        self.window.update_selection(
-            [item.obj for item in self.views[i].selectedItems()])
-
-    def update_selection_0(self):
-        if not self._ignore_selection_changes:
-            self._ignore_selection_changes = True
-            self.update_selection(0, 1)
-            self._ignore_selection_changes = False
-
-    def update_selection_1(self):
-        if not self._ignore_selection_changes:
-            self._ignore_selection_changes = True
-            self.update_selection(1, 0)
-            self._ignore_selection_changes = False
+    def _update_selection(self, selected_view):
+        for view in self._views:
+            if view != selected_view:
+                view.clearSelection()
+            else:
+                self._selected_view = view
+                self.window.update_selection([item.obj for item in view.selectedItems()])
 
     def update_current_view(self):
-        self.update_selection(self._selected_view, abs(self._selected_view - 1))
+        self._update_selection(self._selected_view)
 
     def remove(self, objects):
         self._ignore_selection_changes = True
         self.tagger.remove(objects)
         self._ignore_selection_changes = False
 
-        view = self.views[self._selected_view]
+        view = self._selected_view
         index = view.currentIndex()
         if index.isValid():
             # select the current index
@@ -284,6 +309,16 @@ class MainPanel(QtWidgets.QSplitter):
         else:
             self.update_current_view()
 
+    def set_sorting(self, sort=True):
+        for view in self._views:
+            view.setSortingEnabled(sort)
+
+    def collapse_clusters(self, collapse=True):
+        if collapse:
+            self._views[0].collapseAll()
+        else:
+            self._views[0].expandAll()
+
 
 def paint_fingerprint_icon(painter, rect, icon):
     if not icon:
@@ -422,6 +457,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
         self.select_all_action.triggered.connect(self.selectAll)
         self.select_all_action.setShortcut(QtGui.QKeySequence(_("Ctrl+A")))
         self.doubleClicked.connect(self.activate_item)
+        self.setUniformRowHeights(True)
 
     def contextMenuEvent(self, event):
         item = self.itemAt(event.pos())
@@ -566,7 +602,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             menu.addSeparator()
 
         # Using type here is intentional. isinstance will return true for the
-        # NatAlbum instance, which can't be part of a collection.
+        # NatAlbum instance, which can't be part of a collection.
         # pylint: disable=C0123
         selected_albums = [a for a in self.window.selected_objects if type(a) == Album]
         if selected_albums:
@@ -838,9 +874,18 @@ class TreeItem(QtWidgets.QTreeWidgetItem):
         if obj is not None:
             obj.item = self
         self.sortable = sortable
-        self.setTextAlignment(1, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
-        self.setTextAlignment(7, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
-        self.setTextAlignment(8, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
+        self._sortkeys = {}
+        for column in [
+            MainPanel.LENGTH_COLUMN,
+            MainPanel.TRACKNUMBER_COLUMN,
+            MainPanel.DISCNUMBER_COLUMN,
+        ]:
+            self.setTextAlignment(column, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
+        self.setSizeHint(MainPanel.TITLE_COLUMN, ICON_SIZE)
+
+    def setText(self, column, text):
+        self._sortkeys[column] = None
+        return super().setText(column, text)
 
     def __lt__(self, other):
         if not self.sortable:
@@ -849,9 +894,18 @@ class TreeItem(QtWidgets.QTreeWidgetItem):
         return self.sortkey(column) < other.sortkey(column)
 
     def sortkey(self, column):
-        if column == 1:
-            return self.obj.metadata.length or 0
-        return natsort.natkey(self.text(column).lower())
+        sortkey = self._sortkeys.get(column)
+        if sortkey is not None:
+            return sortkey
+
+        if column == MainPanel.LENGTH_COLUMN:
+            sortkey = self.obj.metadata.length or 0
+        elif column in MainPanel.NAT_SORT_COLUMNS:
+            sortkey = natsort.natkey(self.text(column).lower())
+        else:
+            sortkey = self.text(column).lower()
+        self._sortkeys[column] = sortkey
+        return sortkey
 
 
 class ClusterItem(TreeItem):
@@ -867,7 +921,7 @@ class ClusterItem(TreeItem):
         if self.obj.special and album and album.loaded:
             album.item.update(update_tracks=False)
         if self.isSelected():
-            TreeItem.window.update_selection()
+            TreeItem.window.update_selection(new_selection=False)
 
     def add_file(self, file):
         self.add_files([file])
@@ -881,8 +935,8 @@ class ClusterItem(TreeItem):
         # Benchmarked performance was not noticeably different.
         for file in files:
             item = FileItem(file, True)
-            item.update()
             self.addChild(item)
+            item.update()
 
     def remove_file(self, file):
         file.item.setSelected(False)
@@ -943,7 +997,7 @@ class AlbumItem(TreeItem):
         for i, column in enumerate(MainPanel.columns):
             self.setText(i, album.column(column[1]))
         if selection_changed:
-            TreeItem.window.panel.update_current_view()
+            TreeItem.window.update_selection(new_selection=False)
         # Workaround for PICARD-1446: Expand/collapse indicator for the release
         # is briefly missing on Windows
         self.emitDataChanged()
@@ -1028,7 +1082,7 @@ class TrackItem(TreeItem):
             self.setForeground(i, color)
             self.setBackground(i, bgcolor)
         if self.isSelected():
-            TreeItem.window.update_selection()
+            TreeItem.window.update_selection(new_selection=False)
         if update_album:
             self.parent().update(update_tracks=False)
 
@@ -1051,8 +1105,7 @@ class FileItem(TreeItem):
                 tree_widget.setItemWidget(self, MainPanel.FINGERPRINT_COLUMN,
                     FingerprintColumnWidget(file=file))
         if self.isSelected():
-            TreeItem.window.update_selection()
-
+            TreeItem.window.update_selection(new_selection=False)
         parent = self.parent()
         if isinstance(parent, TrackItem) and update_track:
             parent.update(update_files=False)
